// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// This Source Code Form is subject to the terms of the X11 (MIT)
// License. If a copy of the X11 License was not distributed
// with this file, You can obtain one at https://opensource.org/licenses/MIT/.

#ifndef VKF_GRAPHS_HPP
#define VKF_GRAPHS_HPP

#define DEFAULT_DB "mushrooms_db"
#define DEFAULT_HOST "localhost"
#define DEFAULT_PASS "toor"
#define DEFAULT_USER "root"

#include <string>
#include <cassert>
#include <vector>

#include "DotPicture.hpp"

#include <pybind11/pybind11.h>

#include <mariadb++/connection.hpp>
#include <mariadb++/statement.hpp>

namespace vkf {
class Graphs : public std::vector<DotPicture> {
    public:
    /** default constructor */
    Graphs()=delete;
    Graphs( const std::string &encoder_table_name,
            const std::string &lattices_table_name,
            const std::string &database = DEFAULT_DB,
            const std::string &url = DEFAULT_HOST,
            const std::string &user = DEFAULT_USER,
            const std::string &pass = DEFAULT_PASS
            ) {
        /**< Make the MySQL connection */
        // set up the account
        mariadb::account_ref acc = mariadb::account::create(url, user, pass);
        // create connection
        mariadb::connection_ref con = mariadb::connection::create(acc);
        // open database
        con->execute("USE " + database);
        // obtain content of table encoder
        mariadb::result_set_ref result = con->query("SELECT * FROM " + encoder_table_name);
        // reading from a result set
        std::string attribute_name; /**< This variable contains the name of an attribute in current row */
        std::string attribute_value; /**< This variable saves the value in current row */
        std::vector<std::string> used_attributes;
        while (result->next()) {
            attribute_name = result->get_string("attribute_name");
            if((std::find(used_attributes.begin(), used_attributes.end(), attribute_name)) == used_attributes.end()) {
                used_attributes.emplace_back(attribute_name);
                this->emplace_back(attribute_name);
                this->back().emplace_back(result->get_string("attribute_value"));
                this->back().AppendString(std::string("digraph " + attribute_name + " {\n"));
            } else /* if attribute exists */ {
                std::size_t index = this->findIndex(attribute_name);
                if (index == this->size()) {
                    throw std::runtime_error("Nonexistent attribute!");
                } else {
                    this->at(index).emplace_back(result->get_string("attribute_value"));
                }
            }
        }
        // obtain content of table orderings
        result = con->query("SELECT * FROM " + lattices_table_name);
        con->~connection();
        acc->~account();
        // reading from a result set
        while (result->next()) {
            attribute_name = result->get_string("attribute_name");
            if ((std::find(used_attributes.begin(), used_attributes.end(), attribute_name)) == used_attributes.end()) {   // if yes
                throw std::runtime_error("Ordering between values of an nonexistent attribute!");
            } else /* if attribute exists */ {
                std::size_t index = this->findIndex(attribute_name);
                if (index == this->size()) {
                    throw std::runtime_error("Nonexistent attribute!");
                } else {
                    this->at(index).AppendString(" " + result->get_string("source_value") + " -> " + result->get_string("target_value") + ";\n");
                }
            }
        }
        for (auto it = this->begin(); it != this->end(); it++) {
            it->AppendString(std::string("}\n"));
        }
    }

    /** ShowPictures function implementation
     * \return pybind11::list of std::strings for dot_pictures  */
    pybind11::list ShowPictures() const {
        pybind11::list result;
        for (auto it = this->begin(); it != this->end(); it++) {
            result.append(it->ShowDotString());
        }
        return result;
    }

    private:
        std::size_t findIndex(std::string &part_name) {
            std::size_t index = this->size();
            for (std::size_t ndx = 0; ndx < index; ndx++) {
                if (this->at(ndx).GetAttributeName() == part_name) {
                    index = ndx;
                    break;
                }
            }
            return index;
        }
};

} // namespace vkf
#endif // VKF_GRAPHS_HPP
